# -*- coding: utf-8 -*-
import sys
from shutil import rmtree
from os.path import join
import unittest
from skoolkittest import SkoolKitTestCase
from skoolkit.skoollib import SkoolGame
from skoolkit.skoolhtml import SkoolWriter, FileInfo
from skoolkit.skoolparser import SkoolParser, RefParser

TEST_SKOOL = 'test-html.skool'
TEST_REF = 'test-html.ref'
ODIR = 'html'
GAMEDIR = 'test'
ASMDIR = 'asm'
REFERENCEDIR = 'reference'
UDGDIR = 'images/udgs'

TABLE_SRC = """(data)
{ =h Col1 | =h Col2 | =h,c2 Cols3+4 }
{ =r2 X   | Y       | Za  | Zb }
{           Y2      | Za2 | =t }
"""

TABLE_HTML = """
<table class="data">
<tr>
<th>Col1</th>
<th>Col2</th>
<th colspan="2">Cols3+4</th>
</tr>
<tr>
<td rowspan="2">X</td>
<td>Y</td>
<td>Za</td>
<td>Zb</td>
</tr>
<tr>
<td>Y2</td>
<td>Za2</td>
<td class="transparent"></td>
</tr>
</table>
"""

class SkoolWriterTest(SkoolKitTestCase):
    def read_file(self, fname):
        f = open(join(ODIR, GAMEDIR, fname), 'r')
        contents = f.read()
        f.close()
        return contents

    def _mock_write_image(self, udgs, img_file, scale=2, mask=False, width=None):
        return

    def _get_writer(self, use_ref=False):
        ref_parser = RefParser()
        if use_ref:
            ref_parser.parse(TEST_REF)
        skool_parser = SkoolParser(TEST_SKOOL, html=True)
        skool = SkoolGame(skool_parser, ref_parser)
        file_info = FileInfo(ODIR, GAMEDIR, False)
        writer = SkoolWriter(skool, file_info)
        writer.write_image = self._mock_write_image
        return writer

    def _test_reference_macro(self, macro, def_link_text, page):
        writer = self._get_writer()
        for link_text in ('', '(test)'):
            for anchor in ('', '#test'):
                output = writer.expand('#%s%s%s' % (macro, anchor, link_text), ASMDIR)
                self.link_equals(output, '../%s/%s.html%s' % (REFERENCEDIR, page, anchor), link_text[1:-1] or def_link_text)

    def link_equals(self, html, href, text):
        self.assertEqual(html, '<a class="link" href="%s">%s</a>' % (href, text))

    def img_equals(self, html, alt, src):
        self.assertEqual(html, '<img alt="%s" src="%s" />' % (alt, src))

    def test_macro_bug(self):
        self._test_reference_macro('BUG', 'bug', 'bugs')

    def test_macro_c(self):
        writer = self._get_writer()
        output = writer.expand('#C', ASMDIR)
        self.assertEqual(output, '&#169;')

    def test_macro_call(self):
        writer = self._get_writer()
        f = lambda arg1, arg2: arg2 * arg1
        writer.test_call = lambda cwd, arg1, arg2: f(arg1, arg2)
        output = writer.expand('#CALL:test_call(10,test)', ASMDIR)
        self.assertEqual(output, f(10, 'test'))

    def test_macro_d(self):
        writer = self._get_writer()
        output = writer.expand('#D24592', ASMDIR)
        self.assertEqual(output, 'The final routine')

    def test_macro_erefs(self):
        writer = self._get_writer()
        output = writer.expand('#EREFS24582', ASMDIR)
        self.assertEqual(output, 'routines at <a class="link" href="24586.html">24586</a> and <a class="link" href="24592.html">24592</a>')

    def test_macro_fact(self):
        self._test_reference_macro('FACT', 'fact', 'facts')

    def test_macro_font(self):
        writer = self._get_writer()

        output = writer.expand('#FONT32768,96', ASMDIR)
        self.img_equals(output, 'font', '../images/font/font.png')

        img_fname = 'font2'
        output = writer.expand('#FONT55584,96,,,0(%s)' % img_fname, ASMDIR)
        self.img_equals(output, img_fname, '../images/font/%s.png' % img_fname)

    def test_macro_link(self):
        writer = self._get_writer()

        link_text = 'bugs'
        output = writer.expand('#LINK:Bugs(%s)' % link_text, ASMDIR)
        self.link_equals(output, '../%s/bugs.html' % REFERENCEDIR, link_text)

        link_text = 'pokes'
        anchor = '#poke1'
        output = writer.expand('#LINK:Pokes%s(%s)' % (anchor, link_text), ASMDIR)
        self.link_equals(output, '../%s/pokes.html%s' % (REFERENCEDIR, anchor), link_text)

    def test_macro_name(self):
        writer = self._get_writer()
        anchor = 'testing'
        output = writer.expand('#NAME%s' % anchor, ASMDIR)
        self.assertEqual(output, '<a name="%s"></a>' % anchor)

    def test_macro_p(self):
        writer = self._get_writer()
        output = writer.expand('#P', ASMDIR)
        self.assertEqual(output, '<br /><br />')

    def test_macro_poke(self):
        self._test_reference_macro('POKE', 'poke', 'pokes')

    def test_macro_pokes(self):
        writer = self._get_writer()
        snapshot = writer.skool.snapshot

        output = writer.expand('#POKES32768,255', ASMDIR)
        self.assertEqual(output, '')
        self.assertEqual(snapshot[32768], 255)

        output = writer.expand('#POKES32768,254,10', ASMDIR)
        self.assertEqual(output, '')
        self.assertEqual(snapshot[32768:32778], [254] * 10)

        output = writer.expand('#POKES32768,253,20,2', ASMDIR)
        self.assertEqual(output, '')
        self.assertEqual(snapshot[32768:32808:2], [253] * 20)

    def test_macro_pops(self):
        writer = self._get_writer()
        addr, byte = 49152, 128
        writer.skool.snapshot[addr] = byte
        writer.skool.push_snapshot('test')
        writer.skool.snapshot[addr] = (byte + 127) % 256
        output = writer.expand('#POPS', ASMDIR)
        self.assertEqual(output, '')
        self.assertEqual(writer.skool.snapshot[addr], byte)

    def test_macro_pushs(self):
        writer = self._get_writer()
        addr, byte = 32768, 64
        writer.skool.snapshot[addr] = byte
        output = writer.expand('#PUSHStest', ASMDIR)
        self.assertEqual(output, '')
        self.assertEqual(writer.skool.snapshot[addr], byte)
        writer.skool.snapshot[addr] = (byte + 127) % 256
        writer.skool.pop_snapshot()
        self.assertEqual(writer.skool.snapshot[addr], byte)

    def test_macro_r(self):
        writer = self._get_writer(True)

        output = writer.expand('#R24576', ASMDIR)
        self.link_equals(output, '24576.html', '24576')

        link_text = 'Testing1'
        output = writer.expand('#R24576(%s)' % link_text, ASMDIR)
        self.link_equals(output, '24576.html', link_text)

        output = writer.expand('#R24576', 'other')
        self.link_equals(output, '../%s/24576.html' % ASMDIR, '24576')

        output = writer.expand('#R24579', ASMDIR)
        self.link_equals(output, '24579.html', '6003')

        output = writer.expand('#R24580', ASMDIR)
        self.link_equals(output, '24579.html#24580', '6004')

        output = writer.expand('#R32768@other', ASMDIR)
        self.link_equals(output, '../other/32768.html', '32768')

        link_text = 'Testing2'
        anchor = 'testing3'
        output = writer.expand('#R32768@other#%s(%s)' % (anchor, link_text), ASMDIR)
        self.link_equals(output, '../other/32768.html#%s' % anchor, link_text)

    def test_macro_refs(self):
        writer = self._get_writer()
        output = writer.expand('#REFS24581', ASMDIR)
        self.assertEqual(output, 'routines at <a class="link" href="24583.html">24583</a>, <a class="link" href="24586.html">24586</a> and <a class="link" href="24592.html">24592</a>')

    def test_macro_reg(self):
        writer = self._get_writer()
        for reg in ("a", "b", "c", "d", "e", "h", "l", "i", "r", "ixl", "ixh", "iyl", "iyh", "b'", "c'", "d'", "e'", "h'", "l'", "bc", "de", "hl", "sp", "ix", "iy", "bc'", "de'", "hl'"):
            output = writer.expand('#REG%s' % reg, ASMDIR)
            self.assertEqual(output, '<span class="register">%s</span>' % reg.upper())

    def test_macro_scr(self):
        writer = self._get_writer()

        output = writer.expand('#SCR', ASMDIR)
        self.img_equals(output, 'scr', '../images/scr/scr.png')

        scr_fname = 'scr2'
        output = writer.expand('#SCR2,0,0,10,10(%s)' % scr_fname, ASMDIR)
        self.img_equals(output, scr_fname, '../images/scr/%s.png' % scr_fname)

    def test_macro_space(self):
        writer = self._get_writer()

        output = writer.expand('#SPACE', ASMDIR)
        self.assertEqual(output, '&#160;')

        num = 10
        output = writer.expand('#SPACE%i' % num, ASMDIR)
        self.assertEqual(output, '&#160;' * num)

    def test_macro_table(self):
        writer = self._get_writer()
        output = writer.expand('#TABLE%s\nTABLE#' % TABLE_SRC, ASMDIR)
        self.assertEqual(output, TABLE_HTML)

    def test_macro_udg(self):
        writer = self._get_writer()

        udg_fname = 'udg32768_2x4'
        output = writer.expand('#UDG32768,2', ASMDIR)
        self.img_equals(output, udg_fname, '../%s/%s.png' % (UDGDIR, udg_fname))

        udg_fname = 'test_udg'
        output = writer.expand('#UDG32768,2,6,1,0:49152,2(%s)' % udg_fname, ASMDIR)
        self.img_equals(output, udg_fname, '../%s/%s.png' % (UDGDIR, udg_fname))

    def test_macro_udgarray(self):
        writer = self._get_writer()

        udg_fname = 'test_udg_array'
        output = writer.expand('#UDGARRAY8;32768-32784(%s)' % udg_fname, ASMDIR)
        self.img_equals(output, udg_fname, '../%s/%s.png' % (UDGDIR, udg_fname))

        udg_fname = 'test_udg_array2'
        output = writer.expand('#UDGARRAY8,56,2,256,0;32768;32769;32770;32771;32772x12(%s)' % udg_fname, ASMDIR)
        self.img_equals(output, udg_fname, '../%s/%s.png' % (UDGDIR, udg_fname))

    def test_macro_udgtable(self):
        writer = self._get_writer()
        output = writer.expand('#UDGTABLE%s\nUDGTABLE#' % TABLE_SRC, ASMDIR)
        self.assertEqual(output, TABLE_HTML)

    def test_html_escape(self):
        # Check that HTML characters from the skool file are escaped
        rmtree(ODIR, True)
        writer = self._get_writer()
        fname = 'test.html'
        writer.write_entry(ASMDIR, writer.skool.entries[24576], fname)
        html = self.read_file(join(ASMDIR, fname))
        self.assertTrue('a &lt;= b and b &gt;= c' in html)
        rmtree(ODIR, True)

    def test_html_no_escape(self):
        # Check that HTML characters from the ref file are not escaped
        rmtree(ODIR, True)
        writer = self._get_writer(True)
        writer.write_bugs()
        html = self.read_file(join(REFERENCEDIR, 'bugs.html'))
        self.assertTrue('<p>Hello</p>' in html)
        rmtree(ODIR, True)

if __name__ == '__main__':
    unittest.main()
